# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtCore
from qtpy.QtCharts import QtCharts

STYLES = {QtCore.Qt.SolidLine: "Solid",
          QtCore.Qt.DotLine: "Dot",
          QtCore.Qt.DashDotLine: "Dash-dot"}


class LineSeries(QtCharts.QLineSeries):

    def __getstate__(self):
        return dict(points=self.pointsVector())

    def __setstate__(self, state):
        self.__init__()
        super().append(state["points"])
        pass


if __name__ == "__main__":
    line = LineSeries()
    line.append(0, 1)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import Union
import logging

import qtawesome as qta
from qtpy import QtCore, QtWidgets, QtGui

from prettyqt import core, widgets, gui
from prettyqt.utils import bidict


DOCK_POSITIONS = bidict(top=QtCore.Qt.TopDockWidgetArea,
                        bottom=QtCore.Qt.BottomDockWidgetArea,
                        left=QtCore.Qt.LeftDockWidgetArea,
                        right=QtCore.Qt.RightDockWidgetArea)

TOOLBAR_AREAS = bidict(left=QtCore.Qt.LeftToolBarArea,
                       right=QtCore.Qt.RightToolBarArea,
                       top=QtCore.Qt.TopToolBarArea,
                       bottom=QtCore.Qt.BottomToolBarArea,
                       all=QtCore.Qt.AllToolBarAreas,
                       none=QtCore.Qt.NoToolBarArea)


class MainWindow(QtWidgets.QMainWindow):
    """
    Class for our mainWindow
    includes all docks, a centralwidget and a toolbar
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setDockOptions(self.AllowTabbedDocks |
                            self.AllowNestedDocks |
                            self.GroupedDragging |
                            self.AnimatedDocks)

    def __getitem__(self, index):
        return self.findChild(QtWidgets.QWidget, index)

    def __getstate__(self):
        icon = gui.Icon(self.windowIcon())
        return dict(central_widget=self.centralWidget(),
                    title=self.windowTitle(),
                    is_maximized=self.isMaximized(),
                    icon=icon if not icon.isNull() else None,
                    size=(self.size().width(), self.size().height()))

    def __setstate__(self, state):
        self.__init__()
        self.title = state["title"]
        self.set_icon(state["icon"])
        if state["central_widget"]:
            self.setCentralWidget(state["central_widget"])
        self.resize(state["size"])
        if state["is_maximized"]:
            self.showMaximized()
        self.resize(*state["size"])
        self.box = self.layout()

    def add_toolbar(self, toolbar, position):
        """adds a toolbar to the mainmenu at specified area

        Valid values for position: "left", "right", "top", "bottom"

        Args:
            toolbar: toolbar to use
            position: position of the toolbar

        Raises:
            ValueError: position does not exist
        """
        if position not in TOOLBAR_AREAS:
            raise ValueError("Position not existing")
        self.addToolBar(position, TOOLBAR_AREAS[position])

    def load_window_state(self):
        settings = core.Settings()
        geom = settings.value("mainwindow.geometry", None)
        state = settings.value("mainwindow.state", None)
        if geom is not None and state is not None:
            try:
                self.restoreGeometry(geom)
                self.restoreState(state)
            except TypeError:
                logging.info("Wrong type for window state. Probably Qt binding switch?")
                pass

    def closeEvent(self, event):
        """
        override, gets executed when app gets closed.
        saves GUI settings
        """
        settings = core.Settings()
        settings.set_value("mainwindow.geometry", self.saveGeometry())
        settings.set_value("mainwindow.state", self.saveState())
        super().closeEvent(event)
        event.accept()

    def set_icon(self, icon: Union[QtGui.QIcon, str, None]):
        """set the icon for the menu

        Args:
            icon: icon to use
        """
        if not icon:
            icon = gui.Icon()
        elif isinstance(icon, str):
            icon = qta.icon(icon, color="lightgray")
        self.setWindowIcon(icon)

    def add_widget_as_dock(self,
                           name: str,
                           title: str,
                           vertical: bool = True,
                           position: str = "left") -> widgets.DockWidget:
        dock_widget = widgets.DockWidget(self, name=name, title=title)
        widget = widgets.Widget()
        widget.id = f"{name}.widget"
        orientation = "vertical" if vertical else "horizontal"
        layout = widgets.BoxLayout(orientation, widget)
        layout.set_margin(0)
        dock_widget.setWidget(widget)
        self.add_dockwidget(dock_widget, position)
        dock_widget.box = layout
        return dock_widget

    def add_dockwidget(self, dockwidget, position):
        position = DOCK_POSITIONS[position]
        self.addDockWidget(QtCore.Qt.DockWidgetArea(position), dockwidget)

    def toggle_fullscreen(self):
        """toggle between fullscreen and regular size
        """
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()


MainWindow.__bases__[0].__bases__ = (widgets.Widget,)


if __name__ == "__main__":
    app = widgets.app()
    form = MainWindow()
    form.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import Optional, Union

import qtawesome as qta
from qtpy import QtCore, QtWidgets, QtGui

from prettyqt import widgets, gui


class BaseDialog(QtWidgets.QDialog):

    def __getitem__(self, index):
        return self.findChild(QtWidgets.QWidget, index)

    def __getstate__(self):
        icon = gui.Icon(self.windowIcon())
        return dict(layout=self.layout(),
                    title=self.windowTitle(),
                    is_maximized=self.isMaximized(),
                    has_sizegrip=self.isSizeGripEnabled(),
                    icon=icon if not icon.isNull() else None,
                    size=(self.size().width(), self.size().height()))

    def __setstate__(self, state):
        self.__init__()
        self.title = state["title"]
        self.set_icon(state["icon"])
        if state["layout"]:
            self.set_layout(state["layout"])
        self.resize(state["size"])
        self.setSizeGripEnabled(state["has_sizegrip"])
        if state["is_maximized"]:
            self.showMaximized()
        self.resize(*state["size"])
        self.box = self.layout()

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_Escape:
            self.close()
        elif e.key() == QtCore.Qt.Key_F11:
            self.showNormal() if self.isMaximized() else self.showMaximized()
        else:
            super().keyPressEvent(e)

    def delete_on_close(self):
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

    def add_widget(self, widget):
        self.box += widget
        return widget

    def set_icon(self, icon: Union[QtGui.QIcon, str, None]):
        """set the icon for the menu

        Args:
            icon: icon to use
        """
        if not icon:
            icon = gui.Icon()
        elif isinstance(icon, str):
            icon = qta.icon(icon, color="lightgray")
        self.setWindowIcon(icon)

    def add_buttonbox(self):
        button_box = widgets.DialogButtonBox.create(ok=self.accept,
                                                    cancel=self.reject)
        self.box += button_box
        return button_box

    def show_blocking(self) -> bool:
        self.show()
        return bool(self.exec_())

    def set_flags(self,
                  minimize: bool = None,
                  maximize: bool = None,
                  close: bool = None,
                  stay_on_top: bool = None,
                  window: bool = None):
        if minimize is not None:
            self.setWindowFlag(QtCore.Qt.WindowMinimizeButtonHint, minimize)
        if maximize is not None:
            self.setWindowFlag(QtCore.Qt.WindowMaximizeButtonHint, maximize)
        if close is not None:
            self.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, close)
        if stay_on_top is not None:
            self.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, stay_on_top)
        if window is not None:
            self.setWindowFlag(QtCore.Qt.Window, window)


BaseDialog.__bases__[0].__bases__ = (widgets.Widget,)


class Dialog(BaseDialog):

    DEFAULT_SIZE = None

    def __init__(self,
                 title: str = "",
                 icon=None,
                 parent=None,
                 delete_on_close: bool = True,
                 layout: Optional[str] = None):
        super().__init__(parent=parent)
        if self.DEFAULT_SIZE:
            self.resize(*self.DEFAULT_SIZE)
        self.title = title
        self.set_icon(icon)
        if delete_on_close:
            self.delete_on_close()
        self.box = None
        if layout in ["horizontal", "vertical"]:
            self.box = widgets.BoxLayout(layout)
            self.set_layout(self.box)


if __name__ == "__main__":
    app = widgets.app()
    widget = Dialog()
    import pickle
    with open("data.pkl", "wb") as jar:
        pickle.dump(widget, jar)
    with open("data.pkl", "rb") as jar:
        widget = pickle.load(jar)
    widget.show_blocking()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtCore

from prettyqt import core


class FileDevice(QtCore.QFileDevice):
    pass


FileDevice.__bases__[0].__bases__ = (core.IODevice,)



# -*- coding: utf-8 -*-

import sip
from qtpy import QtCore, QtGui
from prettyqt import constants


class ModelTest(QtCore.QObject):
    def __init__(self, _model, verbose=True, parent=None):
        """
        Connect to all of the models signals,
        Whenever anything happens recheck everything.
        """
        super().__init__(parent)
        self._model = _model
        self.model = sip.cast(_model, QtCore.QAbstractItemModel)
        self.insert = []
        self.remove = []
        self.fetchingMore = False
        self._verbose = verbose
        assert self.model

        self.model.columnsAboutToBeInserted.connect(self.run_all_tests)
        self.model.columnsAboutToBeRemoved.connect(self.run_all_tests)
        self.model.columnsInserted.connect(self.run_all_tests)
        self.model.columnsRemoved.connect(self.run_all_tests)
        self.model.dataChanged.connect(self.run_all_tests)
        self.model.headerDataChanged.connect(self.run_all_tests)
        self.model.layoutAboutToBeChanged.connect(self.run_all_tests)
        self.model.layoutChanged.connect(self.run_all_tests)
        self.model.modelReset.connect(self.run_all_tests)
        self.model.rowsAboutToBeInserted.connect(self.run_all_tests)
        self.model.rowsAboutToBeRemoved.connect(self.run_all_tests)
        self.model.rowsInserted.connect(self.run_all_tests)
        self.model.rowsRemoved.connect(self.run_all_tests)

        # Special checks for inserting/removing
        self.model.rowsAboutToBeInserted.connect(self.rowsAboutToBeInserted)
        self.model.rowsAboutToBeRemoved.connect(self.rowsAboutToBeRemoved)
        self.model.rowsInserted.connect(self.rowsInserted)
        self.model.rowsRemoved.connect(self.rowsRemoved)
        self.run_all_tests()

    def nonDestructiveBasicTest(self):
        """
        nonDestructiveBasicTest tries to call a number of the basic functions (not all)
        to make sure the model doesn't outright segfault,
        testing the functions that makes sense.
        """
        assert(self.model.buddy(QtCore.QModelIndex()) == QtCore.QModelIndex())
        self.model.canFetchMore(QtCore.QModelIndex())
        assert(self.model.columnCount(QtCore.QModelIndex()) >= 0)
        assert(self.model.data(QtCore.QModelIndex(),
                               QtCore.Qt.DisplayRole) == QtCore.QVariant())
        self.fetchingMore = True
        self.model.fetchMore(QtCore.QModelIndex())
        self.fetchingMore = False
        flags = self.model.flags(QtCore.QModelIndex())
        assert(int(flags & constants.IS_ENABLED) == constants.IS_ENABLED or
               int(flags & constants.IS_ENABLED) == 0)
        self.model.hasChildren(QtCore.QModelIndex())
        self.model.hasIndex(0, 0)
        self.model.headerData(0, constants.HORIZONTAL, QtCore.Qt.DisplayRole)
        self.model.index(0, 0, QtCore.QModelIndex())
        self.model.itemData(QtCore.QModelIndex())
        cache = QtCore.QVariant()
        self.model.match(QtCore.QModelIndex(), -1, cache)
        self.model.mimeTypes()
        assert(self.model.parent(QtCore.QModelIndex()) == QtCore.QModelIndex())
        assert(self.model.rowCount(QtCore.QModelIndex()) >= 0)
        variant = QtCore.QVariant()
        self.model.setData(QtCore.QModelIndex(), variant, -1)
        self.model.setHeaderData(-1, constants.HORIZONTAL, QtCore.QVariant())
        self.model.setHeaderData(0, constants.HORIZONTAL, QtCore.QVariant())
        self.model.setHeaderData(999999, constants.HORIZONTAL, QtCore.QVariant())
        self.model.sibling(0, 0, QtCore.QModelIndex())
        self.model.span(QtCore.QModelIndex())
        self.model.supportedDropActions()

    def rowCount(self):
        """
        Tests self.model's implementation of QtCore.QAbstractItemModel::rowCount()
        and hasChildren()

        self.models that are dynamically populated are not as fully tested here.
        """
        # check top row
        topindex = self.model.index(0, 0, QtCore.QModelIndex())
        rows = self.model.rowCount(topindex)
        assert(rows >= 0)
        if rows > 0:
            hasChildren = self.model.hasChildren(topindex)
            assert(hasChildren is True)

        secondlvl = self.model.index(0, 0, topindex)
        if secondlvl.isValid():
            # check a row count where parent is valid
            rows = self.model.rowCount(secondlvl)
            assert(rows >= 0)
            if rows > 0:
                assert(self.model.hasChildren(secondlvl) is True)

        # The self.models rowCount() is tested more extensively in checkChildren,
        # but this catches the big mistakes

    def columnCount(self):
        """
        Tests self.model's implementation of QtCore.QAbstractItemModel::columnCount()
        and hasChildren()
        """
        # check top row
        topidx = self.model.index(0, 0, QtCore.QModelIndex())
        assert(self.model.columnCount(topidx) >= 0)

        # check a column count where parent is valid
        childidx = self.model.index(0, 0, topidx)
        if childidx.isValid():
            assert(self.model.columnCount(childidx) >= 0)

        # columnCount() is tested more extensively in checkChildren,
        # but this catches the big mistakes

    def hasIndex(self):
        """
        Tests self.model's implementation of QtCore.QAbstractItemModel::hasIndex()
        """
        # Make sure that invalid values returns an invalid index
        assert(self.model.hasIndex(-2, -2) is False)
        assert(self.model.hasIndex(-2, 0) is False)
        assert(self.model.hasIndex(0, -2) is False)

        rows = self.model.rowCount(QtCore.QModelIndex())
        cols = self.model.columnCount(QtCore.QModelIndex())

        # check out of bounds
        assert(self.model.hasIndex(rows, cols) is False)
        assert(self.model.hasIndex(rows + 1, cols + 1) is False)

        if rows > 0:
            assert(self.model.hasIndex(0, 0) is True)

        # hasIndex() is tested more extensively in checkChildren()
        # but this catches the big mistakes

    def index(self):
        """
        Tests self.model's implementation of QtCore.QAbstractItemModel::index()
        """
        # Make sure that invalid values returns an invalid index
        assert(self.model.index(-2, -2, QtCore.QModelIndex()) == QtCore.QModelIndex())
        assert(self.model.index(-2, 0, QtCore.QModelIndex()) == QtCore.QModelIndex())
        assert(self.model.index(0, -2, QtCore.QModelIndex()) == QtCore.QModelIndex())

        rows = self.model.rowCount(QtCore.QModelIndex())
        cols = self.model.columnCount(QtCore.QModelIndex())

        if rows == 0:
            return

        # Catch off by one errors
        assert(self.model.index(rows, cols, QtCore.QModelIndex()) == QtCore.QModelIndex())
        assert(self.model.index(0, 0, QtCore.QModelIndex()).isValid() is True)

        # Make sure that the same index is *always* returned
        a = self.model.index(0, 0, QtCore.QModelIndex())
        b = self.model.index(0, 0, QtCore.QModelIndex())
        assert(a == b)

        # index() is tested more extensively in checkChildren()
        # but this catches the big mistakes

    def parent(self):
        """
        Tests self.model's implementation of QtCore.QAbstractItemModel::parent()
        """
        # Make sure the self.model wont crash and will return an invalid QModelIndex
        # when asked for the parent of an invalid index
        assert(self.model.parent(QtCore.QModelIndex()) == QtCore.QModelIndex())

        if self.model.rowCount(QtCore.QModelIndex()) == 0:
            return

        # Column 0              | Column 1  |
        # QtCore.Qself.modelIndex()         |           |
        #    \- topidx          | topidx1   |
        #         \- childix    | childidx1 |

        # Common error test #1, make sure that a top level index has a parent
        # that is an invalid QtCore.Qself.modelIndex
        topidx = self.model.index(0, 0, QtCore.QModelIndex())
        assert(self.model.parent(topidx) == QtCore.QModelIndex())

        # Common error test #2, make sure that a second level index has a parent
        # that is the first level index
        if self.model.rowCount(topidx) > 0:
            childidx = self.model.index(0, 0, topidx)
            assert(self.model.parent(childidx) == topidx)

        # Common error test #3, the second column should NOT have the same children
        # as the first column in a row
        # Usually the second column shouldn't have children
        topidx1 = self.model.index(0, 1, QtCore.QModelIndex())
        if self.model.rowCount(topidx1) > 0:
            childidx = self.model.index(0, 0, topidx)
            childidx1 = self.model.index(0, 0, topidx1)
            assert(childidx != childidx1)

        # Full test, walk n levels deep through the self.model making sure that all
        # parent's children correctly specify their parent
        self.checkChildren(QtCore.QModelIndex())

    def data(self):
        """
        Tests self.model's implementation of QtCore.QAbstractItemModel::data()
        """
        # Invalid index should return an invalid qvariant
        qvar = self.model.data(QtCore.QModelIndex(), QtCore.Qt.DisplayRole)
        assert(qvar is None)

        if self.model.rowCount(QtCore.QModelIndex()) == 0:
            return

        # A valid index should have a valid QtCore.QVariant data
        assert(self.model.index(0, 0, QtCore.QModelIndex()).isValid())

        # shouldn't be able to set data on an invalid index
        assert(self.model.setData(QtCore.QModelIndex(),
                                  QtCore.QVariant("foo"), QtCore.Qt.DisplayRole) is False)

        # General Purpose roles that should return a QString
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               constants.TOOLTIP_ROLE)
        assert data is None or isinstance(data, str)
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               QtCore.Qt.StatusTipRole)
        assert data is None or isinstance(data, str)
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               QtCore.Qt.WhatsThisRole)
        assert data is None or isinstance(data, str)

        # General Purpose roles that should return a QSize
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               QtCore.Qt.SizeHintRole)
        assert data is None or isinstance(data, QtCore.QSize)

        # General Purpose roles that should return a QFont
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               QtCore.Qt.FontRole)
        assert data is None or isinstance(data, QtGui.QFont)

        # Check that the alignment is one we know about
        alignment = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                                    constants.ALIGNMENT_ROLE)
        if alignment is not None:
            flag = QtCore.Qt.AlignHorizontal_Mask | QtCore.Qt.AlignVertical_Mask
            assert(alignment == (alignment & int(flag)))

        # General Purpose roles that should return a QColor
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               constants.BACKGROUND_ROLE)
        assert data is None or isinstance(data, QtGui.QColor)
        data = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                               QtCore.Qt.TextColorRole)
        assert data is None or isinstance(data, QtGui.QColor)

        # Check that the "check state" is one we know about.
        state = self.model.data(self.model.index(0, 0, QtCore.QModelIndex()),
                                QtCore.Qt.CheckStateRole)
        assert state in [None, QtCore.Qt.Unchecked, QtCore.Qt.PartiallyChecked,
                         QtCore.Qt.Checked]

    def run_all_tests(self):
        if self.fetchingMore:
            return

        self.nonDestructiveBasicTest()
        print("passed nonDestructiveBasicTest") if self._verbose else None

        self.rowCount()
        print("passed rowCount") if self._verbose else None

        self.columnCount()
        print("passed columnCount") if self._verbose else None

        self.hasIndex()
        print("passed hasIndex") if self._verbose else None

        self.index()
        print("passed index") if self._verbose else None

        self.parent()
        print("passed parent") if self._verbose else None

        self.data()
        print("passed data") if self._verbose else None
        print("------------------------------") if self._verbose else None

    def rowsAboutToBeInserted(self, parent, start, end):
        """
        Store what is about to be inserted to make sure it actually happens
        """
        c = {"parent": parent, "oldSize": self.model.rowCount(parent),
             "last": self.model.data(self.model.index(start - 1, 0, parent)),
             "next": self.model.data(self.model.index(start, 0, parent))}
        self.insert.append(c)

    def rowsInserted(self, parent, start, end):
        """
        Confirm that what was said was going to happen actually did
        """
        c = self.insert.pop()
        assert(c["parent"] == parent)
        assert(c["oldSize"] + (end - start + 1) == self.model.rowCount(parent))
        assert(c["last"] == self.model.data(self.model.index(start - 1, 0, c["parent"])))

        # if c['next'] != self.model.data(model.index(end+1, 0, c['parent'])):
        #   qDebug << start << end
        #   for i in range(0, self.model.rowCount(QtCore.QModelIndex())):
        #       qDebug << self.model.index(i, 0).data().toString()
        #   qDebug() << c['next'] << self.model.data(model.index(end+1, 0, c['parent']))

        assert(c["next"] == self.model.data(self.model.index(end + 1, 0, c["parent"])))

    def rowsAboutToBeRemoved(self, parent, start, end):
        """
        Store what is about to be inserted to make sure it actually happens
        """
        c = {"parent": parent, "oldSize": self.model.rowCount(parent),
             "last": self.model.data(self.model.index(start - 1, 0, parent)),
             "next": self.model.data(self.model.index(end + 1, 0, parent))}
        self.remove.append(c)

    def rowsRemoved(self, parent, start, end):
        """
        Confirm that what was said was going to happen actually did
        """
        c = self.remove.pop()
        assert(c["parent"] == parent)
        assert(c["oldSize"] - (end - start + 1) == self.model.rowCount(parent))
        assert(c["last"] == self.model.data(self.model.index(start - 1, 0, c["parent"])))
        assert(c["next"] == self.model.data(self.model.index(start, 0, c["parent"])))

    def checkChildren(self, parent, depth=0):
        """
        Called from parent() test.

        A self.model that returns an index of parent X should also return X when asking
        for the parent of the index

        This recursive function does pretty extensive testing on the whole self.model in
        an effort to catch edge cases.

        This function assumes that rowCount(QtCore.QModelIndex()),
        columnCount(QtCore.QModelIndex()) and index() already work.
        If they have a bug it will point it out, but the above tests should have already
        found the basic bugs because it is easier to figure out the problem in
        those tests then this one
        """
        # First just try walking back up the tree.
        p = parent
        while p.isValid():
            p = p.parent()

        # For self.models that are dynamically populated
        if self.model.canFetchMore(parent):
            self.fetchingMore = True
            self.model.fetchMore(parent)
            self.fetchingMore = False

        rows = self.model.rowCount(parent)
        cols = self.model.columnCount(parent)

        if rows > 0:
            assert(self.model.hasChildren(parent))

        # Some further testing against rows(), columns, and hasChildren()
        assert(rows >= 0)
        assert(cols >= 0)

        if rows > 0:
            assert(self.model.hasChildren(parent) is True)

        # qDebug() << "parent:" << self.model.data(parent).toString() << "rows:" << rows
        #          << "columns:" << cols << "parent column:" << parent.column()

        assert(self.model.hasIndex(rows + 1, 0, parent) is False)
        for r in range(0, rows):
            if self.model.canFetchMore(parent):
                self.fetchingMore = True
                self.model.fetchMore(parent)
                self.fetchingMore = False
            assert(self.model.hasIndex(r, cols + 1, parent) is False)
            for c in range(0, cols):
                assert(self.model.hasIndex(r, c, parent))
                index = self.model.index(r, c, parent)
                # rowCount(QtCore.QModelIndex()) and columnCount(QtCore.QModelIndex())
                # said that it existed...
                assert(index.isValid() is True)

                # index() should always return the same index when called twice in a row
                modIdx = self.model.index(r, c, parent)
                assert(index == modIdx)

                # Make sure we get the same index if we request it twice in a row
                a = self.model.index(r, c, parent)
                b = self.model.index(r, c, parent)
                assert(a == b)

                # Some basic checking on the index that is returned
                assert(index.model() == self._model)
                assert(index.row() == r)
                assert(index.column() == c)
                # While you can technically return a QtCore.QVariant usually this is a
                # sign if a bug in data() Disable if this really is ok in your self.model
                assert(self.model.data(index, QtCore.Qt.DisplayRole) is not None)

                # if the next test fails here is some somehwat useful debug you play with
                # if self.model.parent(index) != parent:
                #   qDebug() << r << c << depth << self.model.data(index).toString()
                #        << self.model.data(parent).toString()
                #   qDebug() << index << parent << self.model.parent(index)
                #   # And a view that you can even use to show the self.model
                #   # view = QtWidgets.QTreeView()
                #   # view.setself.model(model)
                #   # view.show()
                #

                # Check that we can get back our real parent
                p = self.model.parent(index)
                assert(p.internalId() == parent.internalId())
                assert(p.row() == parent.row())

                # recursively go down the children
                if self.model.hasChildren(index) and depth < 10:
                    # qDebug() << r << c << "hasChildren" << self.model.rowCount(index)
                    depth += 1
                    self.checkChildren(index, depth)
                # else:
                #   if depth >= 10:
                #       qDebug() << "checked 10 deep"

                # Make sure that after testing the children that the index doesn't change
                newIdx = self.model.index(r, c, parent)
                assert(index == newIdx)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `prettyqt` package."""

import pytest

from prettyqt import syntaxhighlighters


def test_jsonhighlighter():
    highlighter = syntaxhighlighters.JsonHighlighter()
    highlighter.highlightBlock('{"a": "b"}')


def test_pythonhighlighter():
    highlighter = syntaxhighlighters.PythonHighlighter()
    highlighter.highlightBlock("def test(): pass")


def test_yamlhighlighter():
    syntaxhighlighters.YamlHighlighter()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import Dict, Callable, Optional
from contextlib import contextmanager

from qtpy import QtWidgets, QtCore, QtGui

from prettyqt import widgets, core
from prettyqt.utils import bidict


POLICIES = bidict(none=QtCore.Qt.NoContextMenu,
                  prevent=QtCore.Qt.PreventContextMenu,
                  default=QtCore.Qt.DefaultContextMenu,
                  actions=QtCore.Qt.ActionsContextMenu,
                  custom=QtCore.Qt.CustomContextMenu,
                  showhide_menu="showhide_menu")

MODALITIES = bidict(window=QtCore.Qt.WindowModal,
                    application=QtCore.Qt.ApplicationModal,
                    none=QtCore.Qt.NonModal)


class Widget(QtWidgets.QWidget):

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.__getstate__()}"

    def __getstate__(self):
        return dict(layout=self.layout(),
                    size_policy=self.get_size_policy(),
                    accessible_name=self.accessibleName(),
                    tooltip=self.toolTip(),
                    statustip=self.statusTip())

    def __setstate__(self, state):
        self.__init__()
        self.set_layout(state["layout"])
        self.setSizePolicy(state["size_policy"])
        self.setAccessibleName(state["accessible_name"])
        self.setToolTip(state.get("tooltip", ""))
        self.setStatusTip(state.get("statustip", ""))

    def resize(self, *size):
        if isinstance(size[0], tuple):
            super().resize(*size[0])
        else:
            super().resize(*size)

    @property
    def title(self) -> str:
        return self.windowTitle()

    @title.setter
    def title(self, name: str):
        self.setWindowTitle(name)

    @property
    def enabled(self) -> bool:
        return self.isEnabled()

    @enabled.setter
    def enabled(self, state: bool):
        self.setEnabled(state)

    def set_enabled(self):
        self.setEnabled(True)

    def set_disabled(self):
        self.setEnabled(False)

    def set_title(self, title: str):
        self.setWindowTitle(title)

    def set_modality(self, modality: str = "window"):
        """set modality for the dialog

        Valid values for modality: "modeless", "window", "application"

        Args:
            modality: modality for the main window (default: {"window"})

        Raises:
            ValueError: modality type does not exist
        """
        if modality not in MODALITIES:
            raise ValueError("Invalid value for modality.")
        self.setWindowModality(MODALITIES[modality])

    def get_modality(self) -> str:
        """get the current modality modes as a string

        Possible values: "modeless", "window", "application"

        Returns:
            modality mode
            str
        """
        return MODALITIES.inv[self.windowModality()]

    @contextmanager
    def updates_off(self):
        self.setUpdatesEnabled(False)
        yield None
        self.setUpdatesEnabled(True)

    def set_size_policy(self,
                        horizontal: Optional[str] = None,
                        vertical: Optional[str] = None):
        """sets the sizes policy

        possible values for both parameters are "fixed", "minimum", "maximum",
        "preferred", "expanding", "minimum_expanding" and "ignored"

        Args:
            horizontal: horizontal size policy
            vertical: vertical size policy
        """
        sp = self.get_size_policy()
        if horizontal is not None:
            sp.set_horizontal_policy(horizontal)
        if vertical is not None:
            sp.set_vertical_policy(vertical)
        self.setSizePolicy(sp)

    def get_size_policy(self):
        qpol = self.sizePolicy()
        if isinstance(qpol, widgets.SizePolicy):
            return qpol
        pol = widgets.SizePolicy(qpol.horizontalPolicy(),
                                 qpol.verticalPolicy(),
                                 qpol.controlType())
        pol.setHeightForWidth(qpol.hasHeightForWidth())
        pol.setWidthForHeight(qpol.hasWidthForHeight())
        pol.setHorizontalStretch(qpol.horizontalStretch())
        pol.setVerticalStretch(qpol.verticalStretch())
        return pol

    def set_color(self, color):
        self.setStyleSheet(f"background-color: {color};")

    def set_stylesheet(self, item, dct: Dict[str, str]) -> str:
        ss = "; ".join(f"{k.replace('_', '-')}: {v}" for k, v in dct.items())
        stylesheet = f"{item} {{{ss};}}"
        self.setStyleSheet(stylesheet)
        return stylesheet

    def set_contextmenu_policy(self, policy: str):
        """set contextmenu policy for given item view

        Allowed values are "none", "prevent", "default", "actions",
                           "custom", "showhide_menu"

        Args:
            policy: contextmenu policy to use

        Raises:
            ValueError: policy does not exist
        """
        if policy not in POLICIES:
            raise ValueError("invalid selection behaviour")
        self.setContextMenuPolicy(POLICIES[policy])

    def get_contextmenu_policy(self) -> str:
        """returns current contextmenu policy

        Possible values: "none", "prevent", "default", "actions",
                         "custom", "showhide_menu"

        Returns:
            contextmenu policy
        """
        return POLICIES.inv[self.contextMenuPolicy()]

    def set_custom_menu(self, method: Callable):
        self.set_contextmenu_policy("custom")
        self.customContextMenuRequested.connect(method)

    def set_layout(self, layout):
        if layout in ["horizontal", "vertical"]:
            self.box = widgets.BoxLayout(layout)
        elif layout == "grid":
            self.box = widgets.GridLayout()
        elif layout == "form":
            self.box = widgets.FormLayout()
        elif layout == "stacked":
            self.box = widgets.StackedLayout()
        elif layout == "flow":
            from prettyqt import custom_widgets
            self.box = custom_widgets.FlowLayout()
        else:
            self.box = layout
        if self.box is not None:
            self.setLayout(self.box)


Widget.__bases__[0].__bases__ = (core.Object, QtGui.QPaintDevice)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = Widget()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtCore

from prettyqt import core


class Translator(QtCore.QTranslator):
    pass


Translator.__bases__[0].__bases__ = (core.Object,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets


class TreeWidget(QtWidgets.QTreeWidget):
    pass


TreeWidget.__bases__[0].__bases__ = (widgets.TreeView,)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = TreeWidget()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict


VIEW_MODES = bidict(list=QtWidgets.QListView.ListMode,
                    icon=QtWidgets.QListView.IconMode)


class ListView(QtWidgets.QListView):

    def set_view_mode(self, mode: str):
        """set view mode

        possible values are "list", "icon"

        Args:
            mode: view mode to use

        Raises:
            ValueError: invalid view mode
        """
        if mode not in VIEW_MODES:
            raise ValueError(f"Invalid value. Valid values: {VIEW_MODES.keys()}")
        self.setViewMode(VIEW_MODES[mode])

    def get_view_mode(self) -> str:
        """returns view mode

        possible values are "list", "icon"

        Returns:
            view mode
        """
        return VIEW_MODES.inv[self.viewMode()]


ListView.__bases__[0].__bases__ = (widgets.AbstractItemView,)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    dlg = ListView()
    dlg.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets


class Frame(QtWidgets.QFrame):
    pass


Frame.__bases__[0].__bases__ = (widgets.Widget,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict


VIEW_MODES = bidict(default=QtWidgets.QMdiArea.SubWindowView,
                    tabbed=QtWidgets.QMdiArea.TabbedView)


WINDOW_ORDERS = bidict(creation=QtWidgets.QMdiArea.CreationOrder,
                       stacking=QtWidgets.QMdiArea.StackingOrder,
                       activation=QtWidgets.QMdiArea.ActivationHistoryOrder)


class MdiArea(QtWidgets.QMdiArea):

    def __add__(self, other):
        if isinstance(other, QtWidgets.QWidget):
            self.add(other)
            return self

    def set_view_mode(self, mode: str):
        """set view mode for the MDI area

        Valid values are "default" and "tabbed"

        Args:
            mode: view mode to use

        Raises:
            ValueError: view mode does not exist
        """
        if mode not in VIEW_MODES:
            raise ValueError("Invalid value for mode.")
        self.setViewMode(VIEW_MODES[mode])

    def get_view_mode(self) -> str:
        return VIEW_MODES.inv[self.viewMode()]

    def set_window_order(self, mode: str):
        """set the window order behaviour for the MDI area

        Valid values are "creation", "stacking", "activation"

        Args:
            mode: window order behaviour to use

        Raises:
            ValueError: window order mode not existing.
        """
        if mode not in WINDOW_ORDERS:
            raise ValueError("Invalid value for mode.")
        self.setViewMode(WINDOW_ORDERS[mode])

    def get_window_order(self) -> str:
        return WINDOW_ORDERS.inv[self.viewMode()]

    def add(self, item: QtWidgets.QWidget):
        if not isinstance(item, QtWidgets.QMdiSubWindow):
            widget = widgets.MdiSubWindow()
            widget.setWidget(item)
            self.addSubWindow(widget)
        else:
            self.addSubWindow(item)


MdiArea.__bases__[0].__bases__ = (widgets.AbstractScrollArea,)


if __name__ == "__main__":
    app = widgets.app()
    widget = MdiArea()
    le = widgets.LineEdit("test")
    le2 = widgets.LineEdit("test")
    widget.add(le)
    widget.add(le2)
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

import pathlib

from qtpy import QtCore


class Dir(QtCore.QDir):

    def __str__(self):
        return self.absolutePath()

    def to_path(self):
        return pathlib.Path(self.absolutePath())



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtGui

from prettyqt import gui, core


class StandardItemModel(QtGui.QStandardItemModel):

    def __getitem__(self, index):
        return self.item(index)

    def __iter__(self):
        return iter(self.get_children())

    def __getstate__(self):
        return dict(items=self.get_children())

    def __setstate__(self, state):
        self.__init__()
        for item in state["items"]:
            self.appendRow(item)

    def __add__(self, other):
        if isinstance(other, (QtGui.QStandardItem, str)):
            self.add(other)
            return self
        raise TypeError("wrong type for addition")

    def get_children(self):
        return [self.item(index) for index in range(self.rowCount())]

    def add(self, item) -> int:
        if isinstance(item, str):
            item = gui.StandardItem(item)
        self.appendRow(item)
        return self.rowCount()


StandardItemModel.__bases__[0].__bases__ = (core.AbstractItemModel,)


if __name__ == "__main__":
    import pickle
    from prettyqt import widgets
    model = gui.StandardItemModel()
    model.add("test")
    app = widgets.app()
    w = widgets.ListView()
    w.setModel(model)
    model += gui.StandardItem("Item")
    for item in model:
        pass
    with open("data.pkl", "wb") as jar:
        pickle.dump(model, jar)
    with open("data.pkl", "rb") as jar:
        model = pickle.load(jar)
    model += gui.StandardItem("Item2")
    w.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtCore, QtWidgets

from prettyqt import core, widgets


class PopupInfo(QtWidgets.QDialog):
    """
    dialog overlay to show some info to user
    """

    def __init__(self, parent=None, text=None):
        super().__init__(parent=parent)
        self.timer = core.Timer.single_shot(callback=self.close)
        self.label = widgets.Label()
        self.setWindowFlags(self.windowFlags() |
                            QtCore.Qt.Tool |
                            QtCore.Qt.WindowStaysOnTopHint |
                            QtCore.Qt.FramelessWindowHint)
        layout = widgets.BoxLayout("vertical")
        layout.set_margin(20)
        self.set_layout(layout)
        self.setStyleSheet("background-color: black;")
        self.label.setStyleSheet("color: white;")
        # self.setStyleSheet("")
        layout += self.label
        # signals.signals.popup_info.connect(self.popup)

    def show(self, *args, **kwargs):
        self.hide()
        screen_geo = widgets.Application.desktop().screenGeometry()
        size = self.label.sizeHint()
        x = (screen_geo.width() - size.width()) / 2
        y = (screen_geo.height() - size.height()) / 2
        self.move(x, y - 200)
        super().show(*args, **kwargs)
        self.timer.start(2500)

    def show_popup(self, text: str):
        self.label.setText(text)
        self.show()


if __name__ == "__main__":
    app = widgets.app()
    widget = PopupInfo()
    widget.show_popup("test")
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

import pathlib
import sys
from typing import Optional, Union

import qtawesome as qta
from qtpy import QtCore, QtWidgets, QtGui

from prettyqt import core, gui


class Application(QtWidgets.QApplication):

    def set_icon(self, icon: Union[QtGui.QIcon, str, None]):
        """set the icon for the menu

        Args:
            icon: icon to use
        """
        if not icon:
            icon = gui.Icon()
        elif isinstance(icon, str):
            icon = qta.icon(icon, color="lightgray")
        self.setWindowIcon(icon)

    def load_language_file(self, path: pathlib.Path):
        translator = core.Translator(self)
        translator.load(str(path))
        self.installTranslator(translator)

    def set_metadata(self,
                     app_name: Optional[str] = None,
                     app_version: Optional[str] = None,
                     org_name: Optional[str] = None,
                     org_domain: Optional[str] = None):
        if app_name:
            self.setApplicationName(app_name)
        if app_version:
            self.setApplicationVersion(app_name)
        if org_name:
            self.setOrganizationName(org_name)
        if org_domain:
            self.setOrganizationDomain(org_domain)

    def main_loop(self):
        return self.exec_()

    @classmethod
    def use_hdpi_bitmaps(cls):
        cls.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

    @classmethod
    def disable_window_help_button(cls):
        cls.setAttribute(QtCore.Qt.AA_DisableWindowContextHelpButton)

    @classmethod
    def copy_to_clipboard(cls, text: str):
        """
        Sets clipboard to supplied text
        """
        cb = cls.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(text, mode=cb.Clipboard)

    @classmethod
    def get_mainwindow(cls) -> Optional[QtWidgets.QMainWindow]:
        widget_list = cls.instance().topLevelWidgets()
        for widget in widget_list:
            if isinstance(widget, QtWidgets.QMainWindow):
                return widget
        return None

    @classmethod
    def create_default_app(cls) -> "Application":
        cls.disable_window_help_button()
        cls.use_hdpi_bitmaps()
        app = cls(sys.argv)
        return app


Application.__bases__[0].__bases__ = (gui.GuiApplication,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets


class WidgetAction(QtWidgets.QWidgetAction):
    pass


WidgetAction.__bases__[0].__bases__ = (widgets.Action,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtGui

from prettyqt import core


class SyntaxHighlighter(QtGui.QSyntaxHighlighter):
    pass


SyntaxHighlighter.__bases__[0].__bases__ = (core.Object,)



# -*- coding: utf-8 -*-

# credits to https://stackoverflow.com/a/54819051

from prettyqt import widgets, core, gui

from qtpy import QtCore


class LabeledSlider(widgets.Widget):

    value_changed = core.Signal(int)

    def __init__(self, labels, orientation=QtCore.Qt.Horizontal, parent=None):
        super().__init__(parent=parent)

        levels = range(len(labels))
        if not isinstance(labels, (tuple, list)):
            raise Exception("<labels> is a list or tuple.")
        self.levels = list(zip(levels, labels))

        if orientation == QtCore.Qt.Horizontal:
            self.layout = widgets.BoxLayout("vertical", self)
        elif orientation == QtCore.Qt.Vertical:
            self.layout = widgets.BoxLayout("horizontal", self)
        else:
            raise Exception("<orientation> wrong.")

        # gives some space to print labels
        self.left_margin = 10
        self.top_margin = 10
        self.right_margin = 10
        self.bottom_margin = 10

        self.layout.setContentsMargins(self.left_margin, self.top_margin,
                                       self.right_margin, self.bottom_margin)

        self.sl = widgets.Slider(orientation, self)
        self.sl.value_changed.connect(self.value_changed)
        self.sl.setMinimum(0)
        self.sl.setMaximum(len(labels) - 1)
        self.sl.setValue(0)
        if orientation == QtCore.Qt.Horizontal:
            self.sl.setTickPosition(widgets.Slider.TicksBelow)
            self.sl.setMinimumWidth(300)  # just to make it easier to read
        else:
            self.sl.setTickPosition(widgets.Slider.TicksLeft)
            self.sl.setMinimumHeight(300)  # just to make it easier to read
        self.sl.setTickInterval(1)
        self.sl.setSingleStep(1)

        self.layout.addWidget(self.sl)

    def paintEvent(self, e):

        super().paintEvent(e)

        style = self.sl.style()
        painter = gui.Painter(self)
        st_slider = widgets.StyleOptionSlider()
        st_slider.initFrom(self.sl)
        st_slider.orientation = self.sl.orientation()

        length = style.pixelMetric(widgets.Style.PM_SliderLength, st_slider, self.sl)
        available = style.pixelMetric(
            widgets.Style.PM_SliderSpaceAvailable, st_slider, self.sl)

        for v, v_str in self.levels:

            # get the size of the label
            rect = painter.drawText(core.Rect(), QtCore.Qt.TextDontPrint, v_str)

            if self.sl.orientation() == QtCore.Qt.Horizontal:
                x_loc = widgets.Style.sliderPositionFromValue(self.sl.minimum(),
                                                              self.sl.maximum(),
                                                              v,
                                                              available)
                # I assume the offset is half the length of slider, therefore
                # + length//2
                x_loc += length // 2

                # left bound of the text = center - half of text width + L_margin
                left = x_loc - rect.width() // 2 + self.left_margin
                bottom = self.rect().bottom()

                # enlarge margins if clipping
                if v == self.sl.minimum():
                    if left <= 0:
                        self.left_margin = rect.width() // 2 - x_loc
                    if self.bottom_margin <= rect.height():
                        self.bottom_margin = rect.height()

                    self.layout.setContentsMargins(self.left_margin,
                                                   self.top_margin, self.right_margin,
                                                   self.bottom_margin)

                if v == self.sl.maximum() and rect.width() // 2 >= self.right_margin:
                    self.right_margin = rect.width() // 2
                    self.layout.setContentsMargins(self.left_margin,
                                                   self.top_margin, self.right_margin,
                                                   self.bottom_margin)

            else:
                y_loc = widgets.Style.sliderPositionFromValue(self.sl.minimum(),
                                                              self.sl.maximum(),
                                                              v,
                                                              available,
                                                              upsideDown=True)

                bottom = y_loc + length // 2 + rect.height() // 2 + self.top_margin - 3
                # there is a 3 px offset that I can't attribute to any metric

                left = self.left_margin - rect.width()
                if left <= 0:
                    self.left_margin = rect.width() + 2
                    self.layout.setContentsMargins(self.left_margin,
                                                   self.top_margin, self.right_margin,
                                                   self.bottom_margin)

            pos = core.Point(left, bottom)
            painter.drawText(pos, v_str)

        return


if __name__ == '__main__':
    app = widgets.Application([])
    frame = widgets.Widget()
    ha = widgets.BoxLayout("horizontal")
    frame.setLayout(ha)

    w = LabeledSlider(labels=["test", "test2", "test3"], orientation=QtCore.Qt.Vertical)

    ha.addWidget(w)
    frame.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import core


class AbstractItemDelegate(QtWidgets.QAbstractItemDelegate):
    pass


AbstractItemDelegate.__bases__[0].__bases__ = (core.Object,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt.utils import bidict


SIZE_POLICIES = bidict(fixed=QtWidgets.QSizePolicy.Fixed,
                       minimum=QtWidgets.QSizePolicy.Minimum,
                       maximum=QtWidgets.QSizePolicy.Maximum,
                       preferred=QtWidgets.QSizePolicy.Preferred,
                       expanding=QtWidgets.QSizePolicy.Expanding,
                       minimum_expanding=QtWidgets.QSizePolicy.MinimumExpanding,
                       ignored=QtWidgets.QSizePolicy.Ignored)

CONTROL_TYPES = bidict(default=QtWidgets.QSizePolicy.DefaultType,
                       buttonbox=QtWidgets.QSizePolicy.ButtonBox,
                       checkbox=QtWidgets.QSizePolicy.CheckBox,
                       combobox=QtWidgets.QSizePolicy.ComboBox,
                       frame=QtWidgets.QSizePolicy.Frame,
                       groupbox=QtWidgets.QSizePolicy.GroupBox,
                       label=QtWidgets.QSizePolicy.Label,
                       line=QtWidgets.QSizePolicy.Line,
                       lineedit=QtWidgets.QSizePolicy.LineEdit,
                       pushbutton=QtWidgets.QSizePolicy.PushButton,
                       radiobutton=QtWidgets.QSizePolicy.RadioButton,
                       slider=QtWidgets.QSizePolicy.Slider,
                       spinbox=QtWidgets.QSizePolicy.SpinBox,
                       tabwidget=QtWidgets.QSizePolicy.TabWidget,
                       toolbutton=QtWidgets.QSizePolicy.ToolButton)


class SizePolicy(QtWidgets.QSizePolicy):

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.__getstate__()}"

    def __getstate__(self):
        return dict(has_height_for_width=self.hasHeightForWidth(),
                    has_width_for_height=self.hasWidthForHeight(),
                    horizontal_stretch=self.horizontalStretch(),
                    vertical_stretch=self.verticalStretch(),
                    horizontal_policy=self.get_horizontal_policy(),
                    vertical_policy=self.get_vertical_policy(),
                    retain_size_when_hidden=self.retainSizeWhenHidden())

    def __setstate__(self, state):
        self.__init__()
        self.setHeightForWidth(state["has_height_for_width"])
        self.setWidthForHeight(state["has_width_for_height"])
        self.setHorizontalStretch(state["horizontal_stretch"])
        self.setVerticalStretch(state["vertical_stretch"])
        self.set_horizontal_policy(state["horizontal_policy"])
        self.set_vertical_policy(state["vertical_policy"])
        self.setRetainSizeWhenHidden(state["retain_size_when_hidden"])

    def get_horizontal_policy(self) -> str:
        """returns size policy

        possible values are "fixed", "minimum", "maximum", "preferred",
        "expanding", "minimum_expanding" and "ignored"

        Returns:
            horizontal size policy
        """
        return SIZE_POLICIES.inv[self.horizontalPolicy()]

    def set_horizontal_policy(self, mode: str):
        """sets the horizontal policy

        possible values are "fixed", "minimum", "maximum", "preferred",
        "expanding", "minimum_expanding" and "ignored"

        Args:
            mode: policy to set
        """
        self.setHorizontalPolicy(SIZE_POLICIES[mode])

    def get_vertical_policy(self) -> str:
        """returns size policy

        possible values are "fixed", "minimum", "maximum", "preferred",
        "expanding", "minimum_expanding" and "ignored"

        Returns:
            vertical size policy
        """
        return SIZE_POLICIES.inv[self.verticalPolicy()]

    def set_vertical_policy(self, mode: str):
        """sets the horizontal policy

        possible values are "fixed", "minimum", "maximum", "preferred",
        "expanding", "minimum_expanding" and "ignored"

        Args:
            mode: policy to set
        """
        self.setVerticalPolicy(SIZE_POLICIES[mode])

    def get_control_type(self) -> str:
        """returns control type

        possible values are "default", "buttonbox", "checkbox", "combobox", "frame",
        "groupbox", "label", "line", "lineedit", "pushbutton", "radiobutton", "slider",
        "spinbox", "tabwidget", "toolbutton"

        Returns:
            control type
        """
        return CONTROL_TYPES.inv[self.controlType()]

    def set_control_type(self, mode: str):
        """sets the control type

        possible values are "default", "buttonbox", "checkbox", "combobox", "frame",
        "groupbox", "label", "line", "lineedit", "pushbutton", "radiobutton", "slider",
        "spinbox", "tabwidget", "toolbutton"

        Args:
            mode: control type to set
        """
        self.setControlType(CONTROL_TYPES[mode])



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from prettyqt import gui


class CompositeValidator(gui.Validator):

    def __repr__(self):
        return f"CompositeValidator({self.validators})"

    def __init__(self, validators=None, parent=None):
        super().__init__(parent)
        self.validators = validators if validators is not None else []

    def __getstate__(self):
        return dict(validators=self.validators)

    def __setstate__(self, state):
        self.__init__()
        self.validators = state.get("validators", [])

    def validate(self, text, pos=0):
        vals = [v.validate(text, pos)[0] for v in self.validators]
        return (min(vals), text, pos)


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt import custom_validators
    val1 = custom_validators.NotEmptyValidator()
    val2 = custom_validators.PathValidator()
    val = CompositeValidator([val1, val2])
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import List

from qtpy import QtCore, QtWidgets

from prettyqt import widgets, core
from prettyqt.utils import bidict


BUTTONS = bidict(cancel=QtWidgets.QDialogButtonBox.Cancel,
                 ok=QtWidgets.QDialogButtonBox.Ok,
                 save=QtWidgets.QDialogButtonBox.Save,
                 open=QtWidgets.QDialogButtonBox.Open,
                 close=QtWidgets.QDialogButtonBox.Close,
                 discard=QtWidgets.QDialogButtonBox.Discard,
                 apply=QtWidgets.QDialogButtonBox.Apply,
                 reset=QtWidgets.QDialogButtonBox.Reset,
                 restore_defaults=QtWidgets.QDialogButtonBox.RestoreDefaults,
                 help=QtWidgets.QDialogButtonBox.Help,
                 save_all=QtWidgets.QDialogButtonBox.SaveAll,
                 yes=QtWidgets.QDialogButtonBox.Yes,
                 yes_to_all=QtWidgets.QDialogButtonBox.YesToAll,
                 no=QtWidgets.QDialogButtonBox.No,
                 no_to_all=QtWidgets.QDialogButtonBox.NoToAll,
                 abort=QtWidgets.QDialogButtonBox.Abort,
                 retry=QtWidgets.QDialogButtonBox.Retry,
                 ignore=QtWidgets.QDialogButtonBox.Ignore)

ROLES = bidict(invalid=QtWidgets.QDialogButtonBox.InvalidRole,
               accept=QtWidgets.QDialogButtonBox.AcceptRole,
               reject=QtWidgets.QDialogButtonBox.RejectRole,
               destructive=QtWidgets.QDialogButtonBox.DestructiveRole,
               action=QtWidgets.QDialogButtonBox.ActionRole,
               help=QtWidgets.QDialogButtonBox.HelpRole,
               yes=QtWidgets.QDialogButtonBox.YesRole,
               no=QtWidgets.QDialogButtonBox.NoRole,
               apply=QtWidgets.QDialogButtonBox.ApplyRole,
               reset=QtWidgets.QDialogButtonBox.ResetRole)

ORIENTATIONS = bidict(horizontal=QtCore.Qt.Horizontal,
                      vertical=QtCore.Qt.Vertical)


class DialogButtonBox(QtWidgets.QDialogButtonBox):

    button_clicked = core.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clicked.connect(self.on_click)

    def __len__(self):
        return len(self.buttons())

    def __getitem__(self, index):
        return self.button(BUTTONS[index])

    def __iter__(self):
        return iter(self.buttons())

    def __contains__(self, item):
        return self[item] is not None

    @classmethod
    def create(cls, **kwargs):
        box = cls()
        for k, v in kwargs.items():
            btn = box.add_button(k)
            btn.clicked.connect(v)
        return box

    def on_click(self, button):
        self.button_clicked.emit(button.objectName())

    def set_horizontal(self):
        self.setOrientation(QtCore.Qt.Horizontal)

    def set_vertical(self):
        self.setOrientation(QtCore.Qt.Vertical)

    def set_orientation(self, orientation: str):
        """set the orientation of the splitter

        Allowed values are "horizontal", "vertical"

        Args:
            mode: orientation for the splitter

        Raises:
            ValueError: orientation does not exist
        """
        if orientation not in ORIENTATIONS:
            raise ValueError(f"{orientation} not a valid orientation.")
        self.setOrientation(ORIENTATIONS[orientation])

    def get_orientation(self) -> str:
        """returns current orientation

        Possible values: "horizontal", "vertical"

        Returns:
            orientation
        """
        return ORIENTATIONS.inv[self.orientation()]

    def add_buttons(self, buttons: List[str]):
        return [self.add_button(btn) for btn in buttons]

    def add_button(self, button: str):
        """add a default button

        Valid arguments: "cancel", "ok", "save", "open", "close",
                         "discard", "apply", "reset", "restore_defaults",
                         "help", "save_all", "yes", "yes_to_all", "no",
                         "no_to_all", "abort", "retry", "ignore"

        Args:
            button: button to add

        Returns:
            created button

        Raises:
            ValueError: Button type not available
        """
        if button not in BUTTONS:
            raise ValueError("button type not available")
        btn = self.addButton(BUTTONS[button])
        btn.setObjectName(button)
        return btn

    def add_accept_button(self, button: QtWidgets.QPushButton):
        btn = self.addButton(button, self.AcceptRole)
        btn.setObjectName(button)
        return btn

    def add_reject_button(self, button: QtWidgets.QPushButton):
        btn = self.addButton(button, self.RejectRole)
        btn.setObjectName(button)
        return btn


DialogButtonBox.__bases__[0].__bases__ = (widgets.Widget,)


if __name__ == "__main__":
    from prettyqt import widgets
    app = widgets.app()
    widget = DialogButtonBox()
    buttons = list(BUTTONS.keys())
    widget.add_buttons(buttons)
    widget.button_clicked.connect(print)
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import gui


class StylePainter(QtWidgets.QStylePainter):
    pass


StylePainter.__bases__[0].__bases__ = (gui.Painter,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

import pathlib

from qtpy import QtCore, QtWidgets

from prettyqt import core


class FileSystemModel(QtWidgets.QFileSystemModel):
    """
    Class to populate a filesystem treeview
    """
    DATA_ROLE = QtCore.Qt.UserRole + 33
    content_type = "files"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setReadOnly(False)
        self.setRootPath(QtCore.QDir.rootPath())

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if role == self.DATA_ROLE:
            path = index.data(self.FilePathRole)
            return pathlib.Path(path)
        return super().data(index, role)

    def yield_child_indexes(self, index):
        if not self.hasChildren(index):
            return None
        path = self.filePath(index)
        flags = self.filter() | QtCore.QDir.NoDotAndDotDot
        it = core.DirIterator(path, flags)
        while it.hasNext():
            yield self.index(it.next())


FileSystemModel.__bases__[0].__bases__ = (core.AbstractItemModel,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtGui, QtCore

SHAPES = dict(arrow=QtCore.Qt.ArrowCursor,
              uparrow=QtCore.Qt.ArrowCursor,
              cross=QtCore.Qt.CrossCursor,
              wait=QtCore.Qt.WaitCursor,
              caret=QtCore.Qt.IBeamCursor)


class Cursor(QtGui.QCursor):

    def set_shape(self, shape: str):
        self.setShape(SHAPES[shape])



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtCore

from prettyqt import core


class ThreadPool(QtCore.QThreadPool):
    pass


ThreadPool.__bases__[0].__bases__ = (core.Object,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets


class ItemDelegate(QtWidgets.QItemDelegate):
    pass


ItemDelegate.__bases__[0].__bases__ = (widgets.AbstractItemDelegate,)



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import Callable

import qtawesome as qta
from qtpy import QtCore, QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import bidict


STYLES = bidict(icon=QtCore.Qt.ToolButtonIconOnly,
                text=QtCore.Qt.ToolButtonTextOnly,
                text_beside_icon=QtCore.Qt.ToolButtonTextBesideIcon,
                text_below_icon=QtCore.Qt.ToolButtonTextUnderIcon)


class ToolBar(QtWidgets.QToolBar):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_icon_size(24)
        self.menu_buttons = list()

    def add_menu_button(self,
                        label: str,
                        icon,
                        menu: QtWidgets.QMenu) -> widgets.ToolButton:
        btn = widgets.ToolButton.for_menu(menu)
        btn.setText(label)
        btn.setToolButtonStyle(self.toolButtonStyle())
        btn.set_icon(icon)
        self.menu_buttons.append(btn)
        self.addWidget(btn)
        return btn

    def set_style(self, style: str):
        if style is None:
            return None
        self.setToolButtonStyle(STYLES.get(style, 0))
        for btn in self.menu_buttons:
            btn.setToolButtonStyle(STYLES.get(style, 0))

    def get_style(self) -> str:
        return STYLES.inv[self.toolButtonStyle()]

    def add_action(self, label: str, icon, callback: Callable, checkable=False):
        if isinstance(icon, str):
            icon = qta.icon(icon)
        action = self.addAction(icon, label, callback)
        if checkable:
            action.setCheckable(True)
        return action

    def set_icon_size(self, size: int):
        self.setIconSize(core.Size(size, size))

    def set_font_size(self, size: int):
        self.setStyleSheet(f"font-size: {size}pt;")


ToolBar.__bases__[0].__bases__ = (widgets.Widget,)


if __name__ == "__main__":
    app = widgets.app()
    widget = ToolBar()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

import pathlib

from qtpy import QtGui


class Pixmap(QtGui.QPixmap):

    @classmethod
    def from_file(cls, path: pathlib.Path):
        with path.open(mode="rb") as f:
            data = f.read()
        # Create widget
        pixmap = cls()
        pixmap.loadFromData(data)
        return pixmap



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from qtpy import QtWidgets

from prettyqt import widgets


class StatusBar(QtWidgets.QStatusBar):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.progress_bar = widgets.ProgressBar()

    def __add__(self, other):
        if isinstance(other, QtWidgets.QAction):
            self.add_action(other)
            return self
        if isinstance(other, QtWidgets.QWidget):
            self.addWidget(other)
            return self

    def setup_default_bar(self):
        # This is simply to show the bar
        self.progress_bar.hide()
        self.progress_bar.setRange(0, 0)
        self.progress_bar.setFixedSize(200, 20)
        self.progress_bar.setTextVisible(False)
        self.addPermanentWidget(self.progress_bar)

    def add_action(self, action):
        self.addAction(action)

    def set_color(self, color):
        self.setStyleSheet(f"background-color: {color};")

    def add_widget(self, widget, permanent=False):
        if permanent:
            self.addPermanentWidget(widget)
        else:
            self.addWidget(widget)


StatusBar.__bases__[0].__bases__ = (widgets.Widget,)


if __name__ == "__main__":
    app = widgets.app()
    dlg = widgets.MainWindow()
    status_bar = StatusBar()
    status_bar.set_color("black")
    label = widgets.Label("test")
    status_bar.addWidget(label)
    status_bar.setup_default_bar()
    dlg.setStatusBar(status_bar)
    dlg.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

import datetime

from qtpy import QtWidgets

from prettyqt import widgets, core


class DateEdit(QtWidgets.QDateEdit):

    value_changed = core.Signal(datetime.datetime)

    def __getstate__(self):
        return dict(calendar_popup=self.calendarPopup(),
                    date=self.get_date(),
                    display_format=self.displayFormat(),
                    range=(self.min_date(), self.max_date()),
                    tooltip=self.toolTip(),
                    statustip=self.statusTip(),
                    enabled=self.isEnabled())

    def __setstate__(self, state):
        self.__init__(state["date"])
        self.setEnabled(state.get("enabled", True))
        self.setDisplayFormat(state["display_format"])
        self.set_range(*state["range"])
        self.setToolTip(state.get("tooltip", ""))
        self.setStatusTip(state.get("statustip", ""))

    def set_value(self, value):
        return self.setDate(value)

    def set_range(self, lower: datetime.date, upper: datetime.date):
        self.setToolTip(f"{lower} <= x <= {upper}")
        self.setDateRange(lower, upper)

    def get_value(self) -> datetime.date:
        return self.get_date()


DateEdit.__bases__[0].__bases__ = (widgets.DateTimeEdit,)


if __name__ == "__main__":
    from prettyqt import widgets
    app = widgets.app()
    widget = DateEdit()
    widget.show()
    app.exec_()



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import Union

import qtawesome as qta
from qtpy import QtWidgets, QtCore, QtGui

from prettyqt import gui
from prettyqt.utils import bidict


STATES = bidict(unchecked=QtCore.Qt.Unchecked,
                partial=QtCore.Qt.PartiallyChecked,
                checked=QtCore.Qt.Checked)


class TreeWidgetItem(QtWidgets.QTreeWidgetItem):

    def __repr__(self):
        return f"TreeWidgetItem()"

    def __getstate__(self):
        return dict(text=self.text(),
                    tooltip=self.toolTip(),
                    statustip=self.statusTip(),
                    checkstate=self.get_checkstate(),
                    icon=gui.Icon(self.icon()) if not self.icon().isNull() else None,
                    data=self.data(QtCore.Qt.UserRole))

    def __setstate__(self, state):
        self.__init__()
        self.setText(state.get("text", ""))
        self.setToolTip(state.get("tooltip", ""))
        self.setStatusTip(state.get("statustip", ""))
        self.setData(QtCore.Qt.UserRole, state["data"])
        self.set_icon(state["icon"])
        self.set_checkstate(state["checkstate"])

    def set_icon(self, icon: Union[QtGui.QIcon, str, None]):
        """set the icon for the action

        Args:
            icon: icon to use
        """
        if not icon:
            icon = gui.Icon()
        elif isinstance(icon, str):
            icon = qta.icon(icon)
        self.setIcon(icon)

    def set_checkstate(self, state: str):
        """set checkstate of the checkbox

        valid values are: unchecked, partial, checked

        Args:
            state: checkstate to use

        Raises:
            ValueError: invalid checkstate
        """
        if state not in STATES:
            raise ValueError("Invalid checkstate.")
        self.setCheckState(STATES[state])

    def get_checkstate(self) -> str:
        """returns checkstate

        possible values are "unchecked", "partial", "checked"

        Returns:
            checkstate
        """
        return STATES.inv[self.checkState()]



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from prettyqt import core, gui


class YamlHighlighter(gui.SyntaxHighlighter):
    """
    Enabled the syntax highlightning for the yaml files.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.rules = []
        self.commentStart = core.RegExp("#")
        self.commentEnd = core.RegExp("\n|\r")
        self.commentFormat = gui.TextCharFormat()
        self.commentFormat.setFontItalic(True)
        self.commentFormat.set_foreground_color("darkgray")
        f = gui.TextCharFormat()
        r = core.RegExp()
        r.setMinimal(True)
        f.setFontWeight(gui.Font.Normal)
        f.set_foreground_color("blue")
        tag_list = ["\\btrue\\b", "\\bfalse\\b"]
        for tag in tag_list:
            r.setPattern(tag)
            self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.setForeground(gui.Color(127, 64, 127))
        r.setPattern(r"\\d+")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.set_foreground_color("darkblue")
        r.setPattern(r"^\s*[_.\w]*\s*:")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.set_foreground_color("darkblue")
        r.setPattern(r":\s*:[_\.\w]*$|:\s*\@[_\.\w]*$")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.setFontWeight(gui.Font.Bold)
        f.set_foreground_color("darkred")
        r.setPattern(r"^\s*-")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.set_foreground_color("darkred")
        r.setPattern(r"^---$")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.set_foreground_color("darkgreen")
        r.setPattern(r"[\[\]\{\}\,]")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.setFontWeight(gui.Font.Normal)
        f.set_foreground_color("magenta")
        r.setPattern(r"\".*\"|\'.*\'")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.setForeground(gui.Color(127, 64, 127))
        r.setPattern(r"\\$\\(.*\\)")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        f.set_foreground_color("lightgray")
        r.setPattern(r"<!DOCTYPE.*>")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))
        r.setPattern(r"<\\?xml.*\\?>")
        self.rules.append((core.RegExp(r), gui.TextCharFormat(f)))



# -*- coding: utf-8 -*-
"""
@author: Philipp Temminghoff
"""

from typing import Optional, Generator, Any, List
import functools
import logging
import operator

from qtpy import QtCore, QtWidgets

from prettyqt import gui, widgets
from prettyqt.utils import bidict


TRIGGERS = bidict(none=QtWidgets.QAbstractItemView.NoEditTriggers,
                  double_click=QtWidgets.QAbstractItemView.DoubleClicked,
                  edit_key=QtWidgets.QAbstractItemView.EditKeyPressed)

SELECTION_BEHAVIOURS = bidict(rows=QtWidgets.QAbstractItemView.SelectRows,
                              columns=QtWidgets.QAbstractItemView.SelectColumns,
                              items=QtWidgets.QAbstractItemView.SelectItems)

SELECTION_MODES = bidict(single=QtWidgets.QAbstractItemView.SingleSelection,
                         extended=QtWidgets.QAbstractItemView.ExtendedSelection,
                         multi=QtWidgets.QAbstractItemView.MultiSelection,
                         none=QtWidgets.QAbstractItemView.NoSelection)


class AbstractItemView(QtWidgets.QAbstractItemView):

    def __len__(self):
        if self.model() is not None:
            return self.model().rowCount()
        return 0

    def selectAll(self):
        """
        Override, we dont want to selectAll for too many items for performance reasons
        """
        if self.model() is None:
            return None
        if self.model().rowCount() * self.model().columnCount() > 1_000_000:
            logging.info("Too many cells to select.")
            return None
        super().selectAll()

    def setModel(self, model):
        """
        delete old selection model explicitely, seems to help with memory usage
        """
        old_sel_model = self.selectionModel()
        super().setModel(model)
        if old_sel_model:
            del old_sel_model

    def toggle_select_all(self):
        """
        select all items from list (deselect when all selected)
        """
        if self.selectionModel() is None:
            return None
        if self.selectionModel().hasSelection():
            self.clearSelection()
        else:
            self.selectAll()

    def set_table_color(self, color):
        self.setStyleSheet(f"QHeaderView::section {{ background-color:{color} }}")

    def current_index(self) -> Optional[int]:
        if self.selectionModel() is None:
            return None
        return self.selectionModel().currentIndex()

    def current_data(self):
        if self.model() is None:
            return None
        return self.current_index().data(QtCore.Qt.UserRole)

    def selected_indexes(self) -> List[QtCore.QModelIndex]:
        """
        returns list of selected indexes in first row
        """
        indexes = (x for x in self.selectedIndexes() if x.column() == 0)
        return sorted(indexes, key=lambda x: x.row())

    def selected_names(self) -> Generator[Any, None, None]:
        """
        returns generator yielding item names
        """
        return (x.data(self.model().NAME_ROLE)
                for x in self.selected_indexes())

    def selected_rows(self) -> Generator[int, None, None]:
        """
        returns generator yielding row nums
        """
        return (x.row() for x in self.selected_indexes())

    def selected_data(self) -> Generator[Any, None, None]:
        """
        returns generator yielding selected userData
        """
        return (x.data(self.model().QtCore.Qt.UserRole)
                for x in self.selected_indexes())

    def setup_dragdrop_move(self):
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(self.DragDrop)
        self.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.setDropIndicatorShown(True)

    def set_edit_triggers(self, *triggers):
        for item in triggers:
            if item not in TRIGGERS:
                raise ValueError("trigger type not available")
        flags = functools.reduce(operator.ior, [TRIGGERS[t] for t in triggers])
        self.setEditTriggers(flags)

    def get_edit_triggers(self) -> list:
        return [k for k, v in TRIGGERS.items() if v & self.editTriggers()]

    def set_selection_behaviour(self, behaviour: str):
        """set selection behaviour for given item view

        Allowed values are "rows", "columns", "items"

        Args:
            behaviour: selection behaviour to use

        Raises:
            ValueError: behaviour does not exist
        """
        if behaviour not in SELECTION_BEHAVIOURS:
            raise ValueError("invalid selection behaviour")
        self.setSelectionBehavior(SELECTION_BEHAVIOURS[behaviour])

    def get_selection_behaviour(self) -> str:
        """returns current selection behaviour

        Possible values: "rows", "columns", "items"

        Returns:
            selection behaviour
        """
        return SELECTION_BEHAVIOURS.inv[self.selectionBehavior()]

    def set_selection_mode(self, mode: str):
        """set selection mode for given item view

        Allowed values are "single", "extended", "multi" or "none"

        Args:
            mode: selection mode to use

        Raises:
            ValueError: mode does not exist
        """
        if mode not in SELECTION_MODES:
            raise ValueError("Format must be either 'single', 'extended',"
                             "'multi' or 'None'")
        self.setSelectionMode(SELECTION_MODES[mode])

    def get_selection_mode(self) -> str:
        """returns current selection mode

        Possible values: "single", "extended", "multi" or "none"

        Returns:
            selection mode
        """
        return SELECTION_MODES.inv[self.selectionMode()]

    def num_selected(self) -> int:
        """returns amount of selected rows

        Returns:
            amount of selected rows
        """
        if self.selectionModel() is None:
            return 0
        return len(self.selectionModel().selectedRows())

    def jump_to_column(self, col_num: int):
        """make sure column at given index is visible

        scrolls to column at given index

        Args:
            col_num: column to scroll to
        """
        if self.model() is None:
            return None
        idx = self.model().index(0, col_num)
        self.scrollTo(idx)

    def scroll_to_top(self):
        """override to use abstractitemview-way of scrolling to top
        """
        self.scrollToTop()

    def scroll_to_bottom(self):
        """override to use abstractitemview-way of scrolling to bottom
        """
        self.scrollToBottom()

    def select_last_row(self):
        idx = self.model().createIndex(self.model().rowCount() - 1, 0)
        self.setCurrentIndex(idx)

    def highlight_when_inactive(self):
        """also highlight items when widget does not have focus
        """
        p = gui.Palette()
        p.highlight_inactive()
        self.setPalette(p)

    def raise_dock(self) -> bool:
        node = self
        while node:
            node = node.parent()
            if isinstance(node, QtWidgets.QDockWidget):
                node.setVisible(True)
                node.raise_()
                return True
        return False


AbstractItemView.__bases__[0].__bases__ = (widgets.AbstractScrollArea,)
